# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2015 ClusterHQ
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Exceptions that can be raised by libzfs_core operations.
"""
from __future__ import absolute_import, division, print_function

import errno
from ._constants import (
    ECHRNG,
    ECKSUM,
    ETIME,
    ZFS_ERR_CHECKPOINT_EXISTS,
    ZFS_ERR_DISCARDING_CHECKPOINT,
    ZFS_ERR_NO_CHECKPOINT,
    ZFS_ERR_DEVRM_IN_PROGRESS,
    ZFS_ERR_VDEV_TOO_BIG,
    ZFS_ERR_WRONG_PARENT,
    ZFS_ERR_RAIDZ_EXPAND_IN_PROGRESS,
    zfs_errno
)


class ZFSError(Exception):
    errno = None
    message = None
    name = None

    def __str__(self):
        if self.name is not None:
            return "[Errno %d] %s: '%s'" % (
                self.errno, self.message, self.name)
        else:
            return "[Errno %d] %s" % (self.errno, self.message)

    def __repr__(self):
        return "%s(%r, %r)" % (
            self.__class__.__name__, self.errno, self.message)


class ZFSGenericError(ZFSError):

    def __init__(self, errno, name, message):
        self.errno = errno
        self.message = message
        self.name = name


class ZFSInitializationFailed(ZFSError):
    message = "Failed to initialize libzfs_core"

    def __init__(self, errno):
        self.errno = errno


class MultipleOperationsFailure(ZFSError):

    def __init__(self, errors, suppressed_count):
        # Use first of the individual error codes
        # as an overall error code.  This is more consistent.
        self.errno = errors[0].errno
        self.errors = errors
        # this many errors were encountered but not placed on the `errors` list
        self.suppressed_count = suppressed_count

    def __str__(self):
        return "%s, %d errors included, %d suppressed" % (
            ZFSError.__str__(self), len(self.errors), self.suppressed_count)

    def __repr__(self):
        return "%s(%r, %r, errors=%r, suppressed=%r)" % (
            self.__class__.__name__, self.errno, self.message, self.errors,
            self.suppressed_count)


class DatasetNotFound(ZFSError):

    """
    This exception is raised when an operation failure can be caused by a
    missing snapshot or a missing filesystem and it is impossible to
    distinguish between the causes.
    """
    errno = errno.ENOENT
    message = "Dataset not found"

    def __init__(self, name):
        self.name = name


class DatasetExists(ZFSError):

    """
    This exception is raised when an operation failure can be caused by an
    existing snapshot or filesystem and it is impossible to distinguish between
    the causes.
    """
    errno = errno.EEXIST
    message = "Dataset already exists"

    def __init__(self, name):
        self.name = name


class NotClone(ZFSError):
    errno = errno.EINVAL
    message = "Filesystem is not a clone, can not promote"

    def __init__(self, name):
        self.name = name


class FilesystemExists(DatasetExists):
    message = "Filesystem already exists"

    def __init__(self, name):
        self.name = name


class FilesystemNotFound(DatasetNotFound):
    message = "Filesystem not found"

    def __init__(self, name):
        self.name = name


class ParentNotFound(ZFSError):
    errno = errno.ENOENT
    message = "Parent not found"

    def __init__(self, name):
        self.name = name


class WrongParent(ZFSError):
    errno = ZFS_ERR_WRONG_PARENT
    message = "Parent dataset is not a filesystem"

    def __init__(self, name):
        self.name = name


class SnapshotExists(DatasetExists):
    message = "Snapshot already exists"

    def __init__(self, name):
        self.name = name


class SnapshotNotFound(DatasetNotFound):
    message = "Snapshot not found"

    def __init__(self, name):
        self.name = name


class SnapshotNotLatest(ZFSError):
    errno = errno.EEXIST
    message = "Snapshot is not the latest"

    def __init__(self, name):
        self.name = name


class SnapshotIsCloned(ZFSError):
    errno = errno.EEXIST
    message = "Snapshot is cloned"

    def __init__(self, name):
        self.name = name


class SnapshotIsHeld(ZFSError):
    errno = errno.EBUSY
    message = "Snapshot is held"

    def __init__(self, name):
        self.name = name


class DuplicateSnapshots(ZFSError):
    errno = errno.EXDEV
    message = "Requested multiple snapshots of the same filesystem"

    def __init__(self, name):
        self.name = name


class SnapshotFailure(MultipleOperationsFailure):
    message = "Creation of snapshot(s) failed for one or more reasons"

    def __init__(self, errors, suppressed_count):
        super(SnapshotFailure, self).__init__(errors, suppressed_count)


class SnapshotDestructionFailure(MultipleOperationsFailure):
    message = "Destruction of snapshot(s) failed for one or more reasons"

    def __init__(self, errors, suppressed_count):
        super(SnapshotDestructionFailure, self).__init__(
            errors, suppressed_count)


class BookmarkExists(ZFSError):
    errno = errno.EEXIST
    message = "Bookmark already exists"

    def __init__(self, name):
        self.name = name


class BookmarkNotFound(ZFSError):
    errno = errno.ENOENT
    message = "Bookmark not found"

    def __init__(self, name):
        self.name = name


class BookmarkMismatch(ZFSError):
    errno = errno.EINVAL
    message = "source is not an ancestor of the new bookmark's dataset"

    def __init__(self, name):
        self.name = name


class BookmarkSourceInvalid(ZFSError):
    errno = errno.EINVAL
    message = "Bookmark source is not a valid snapshot or existing bookmark"

    def __init__(self, name):
        self.name = name


class BookmarkNotSupported(ZFSError):
    errno = errno.ENOTSUP
    message = "Bookmark feature is not supported"

    def __init__(self, name):
        self.name = name


class BookmarkFailure(MultipleOperationsFailure):
    message = "Creation of bookmark(s) failed for one or more reasons"

    def __init__(self, errors, suppressed_count):
        super(BookmarkFailure, self).__init__(errors, suppressed_count)


class BookmarkDestructionFailure(MultipleOperationsFailure):
    message = "Destruction of bookmark(s) failed for one or more reasons"

    def __init__(self, errors, suppressed_count):
        super(BookmarkDestructionFailure, self).__init__(
            errors, suppressed_count)


class BadHoldCleanupFD(ZFSError):
    errno = errno.EBADF
    message = "Bad file descriptor as cleanup file descriptor"


class HoldExists(ZFSError):
    errno = errno.EEXIST
    message = "Hold with a given tag already exists on snapshot"

    def __init__(self, name):
        self.name = name


class HoldNotFound(ZFSError):
    errno = errno.ENOENT
    message = "Hold with a given tag does not exist on snapshot"

    def __init__(self, name):
        self.name = name


class HoldFailure(MultipleOperationsFailure):
    message = "Placement of hold(s) failed for one or more reasons"

    def __init__(self, errors, suppressed_count):
        super(HoldFailure, self).__init__(errors, suppressed_count)


class HoldReleaseFailure(MultipleOperationsFailure):
    message = "Release of hold(s) failed for one or more reasons"

    def __init__(self, errors, suppressed_count):
        super(HoldReleaseFailure, self).__init__(errors, suppressed_count)


class SnapshotMismatch(ZFSError):
    errno = errno.ENODEV
    message = "Snapshot is not descendant of source snapshot"

    def __init__(self, name):
        self.name = name


class StreamMismatch(ZFSError):
    errno = errno.ENODEV
    message = "Stream is not applicable to destination dataset"

    def __init__(self, name):
        self.name = name


class DestinationModified(ZFSError):
    errno = errno.ETXTBSY
    message = "Destination dataset has modifications that can not be undone"

    def __init__(self, name):
        self.name = name


class BadStream(ZFSError):
    errno = ECKSUM
    message = "Bad backup stream"


class StreamFeatureNotSupported(ZFSError):
    errno = errno.ENOTSUP
    message = "Stream contains unsupported feature"


class UnknownStreamFeature(ZFSError):
    errno = errno.ENOTSUP
    message = "Unknown feature requested for stream"


class StreamFeatureInvalid(ZFSError):
    errno = errno.EINVAL
    message = "Kernel modules must be upgraded to receive this stream"


class StreamFeatureIncompatible(ZFSError):
    errno = errno.EINVAL
    message = "Incompatible embedded feature with encrypted receive"


class StreamTruncated(ZFSError):
    errno = zfs_errno.ZFS_ERR_STREAM_TRUNCATED
    message = "incomplete stream"


class ReceivePropertyFailure(MultipleOperationsFailure):
    message = "Receiving of properties failed for one or more reasons"

    def __init__(self, errors, suppressed_count):
        super(ReceivePropertyFailure, self).__init__(errors, suppressed_count)


class StreamIOError(ZFSError):
    message = "I/O error while writing or reading stream"

    def __init__(self, errno):
        self.errno = errno


class ZIOError(ZFSError):
    errno = errno.EIO
    message = "I/O error"

    def __init__(self, name):
        self.name = name


class NoSpace(ZFSError):
    errno = errno.ENOSPC
    message = "No space left"

    def __init__(self, name):
        self.name = name


class QuotaExceeded(ZFSError):
    errno = errno.EDQUOT
    message = "Quota exceeded"

    def __init__(self, name):
        self.name = name


class DatasetBusy(ZFSError):
    errno = errno.EBUSY
    message = "Dataset is busy"

    def __init__(self, name):
        self.name = name


class NameTooLong(ZFSError):
    errno = errno.ENAMETOOLONG
    message = "Dataset name is too long"

    def __init__(self, name):
        self.name = name


class NameInvalid(ZFSError):
    errno = errno.EINVAL
    message = "Invalid name"

    def __init__(self, name):
        self.name = name


class SnapshotNameInvalid(NameInvalid):
    message = "Invalid name for snapshot"

    def __init__(self, name):
        self.name = name


class FilesystemNameInvalid(NameInvalid):
    message = "Invalid name for filesystem or volume"

    def __init__(self, name):
        self.name = name


class BookmarkNameInvalid(NameInvalid):
    message = "Invalid name for bookmark"

    def __init__(self, name):
        self.name = name


class ReadOnlyPool(ZFSError):
    errno = errno.EROFS
    message = "Pool is read-only"

    def __init__(self, name):
        self.name = name


class SuspendedPool(ZFSError):
    errno = errno.EAGAIN
    message = "Pool is suspended"

    def __init__(self, name):
        self.name = name


class PoolNotFound(ZFSError):
    errno = errno.EXDEV
    message = "No such pool"

    def __init__(self, name):
        self.name = name


class PoolsDiffer(ZFSError):
    errno = errno.EXDEV
    message = "Source and target belong to different pools"

    def __init__(self, name):
        self.name = name


class FeatureNotSupported(ZFSError):
    errno = errno.ENOTSUP
    message = "Feature is not supported in this version"

    def __init__(self, name):
        self.name = name


class PropertyNotSupported(ZFSError):
    errno = errno.ENOTSUP
    message = "Property is not supported in this version"

    def __init__(self, name):
        self.name = name


class PropertyInvalid(ZFSError):
    errno = errno.EINVAL
    message = "Invalid property or property value"

    def __init__(self, name):
        self.name = name


class DatasetTypeInvalid(ZFSError):
    errno = errno.EINVAL
    message = "Specified dataset type is unknown"

    def __init__(self, name):
        self.name = name


class UnknownCryptCommand(ZFSError):
    errno = errno.EINVAL
    message = "Specified crypt command is invalid"

    def __init__(self, name):
        self.name = name


class EncryptionKeyNotLoaded(ZFSError):
    errno = errno.EACCES
    message = "Encryption key is not currently loaded"


class EncryptionKeyAlreadyLoaded(ZFSError):
    errno = errno.EEXIST
    message = "Encryption key is already loaded"


class EncryptionKeyInvalid(ZFSError):
    errno = errno.EACCES
    message = "Incorrect encryption key provided"


class ZCPError(ZFSError):
    errno = None
    message = None


class ZCPSyntaxError(ZCPError):
    errno = errno.EINVAL
    message = "Channel program contains syntax errors"

    def __init__(self, details):
        self.details = details


class ZCPRuntimeError(ZCPError):
    errno = ECHRNG
    message = "Channel programs encountered a runtime error"

    def __init__(self, details):
        self.details = details


class ZCPLimitInvalid(ZCPError):
    errno = errno.EINVAL
    message = "Channel program called with invalid limits"


class ZCPTimeout(ZCPError):
    errno = ETIME
    message = "Channel program timed out"


class ZCPSpaceError(ZCPError):
    errno = errno.ENOSPC
    message = "Channel program exhausted the memory limit"


class ZCPMemoryError(ZCPError):
    errno = errno.ENOMEM
    message = "Channel program return value too large"


class ZCPPermissionError(ZCPError):
    errno = errno.EPERM
    message = "Channel programs must be run as root"


class CheckpointExists(ZFSError):
    errno = ZFS_ERR_CHECKPOINT_EXISTS
    message = "Pool already has a checkpoint"


class CheckpointNotFound(ZFSError):
    errno = ZFS_ERR_NO_CHECKPOINT
    message = "Pool does not have a checkpoint"


class CheckpointDiscarding(ZFSError):
    errno = ZFS_ERR_DISCARDING_CHECKPOINT
    message = "Pool checkpoint is being discarded"


class DeviceRemovalRunning(ZFSError):
    errno = ZFS_ERR_DEVRM_IN_PROGRESS
    message = "A vdev is currently being removed"


class DeviceTooBig(ZFSError):
    errno = ZFS_ERR_VDEV_TOO_BIG
    message = "One or more top-level vdevs exceed the maximum vdev size"


class RaidzExpansionRunning(ZFSError):
    errno = ZFS_ERR_RAIDZ_EXPAND_IN_PROGRESS
    message = "A raidz device is currently expanding"


# vim: softtabstop=4 tabstop=4 expandtab shiftwidth=4



# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2015 ClusterHQ
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Important `libzfs_core` constants.
"""

from __future__ import absolute_import, division, print_function
import errno
import sys


# Compat for platform-specific errnos
if sys.platform.startswith('freebsd'):
    ECHRNG = errno.ENXIO
    ECKSUM = 97  # EINTEGRITY
    ETIME = errno.ETIMEDOUT
else:
    ECHRNG = errno.ECHRNG
    ECKSUM = errno.EBADE
    ETIME = errno.ETIME


# https://stackoverflow.com/a/1695250
def enum_with_offset(offset, sequential, named):
    enums = dict(((b, a + offset) for a, b in enumerate(sequential)), **named)
    return type('Enum', (), enums)


def enum(*sequential, **named):
    return enum_with_offset(0, sequential, named)


#: Maximum length of any ZFS name.
MAXNAMELEN = 255
#: Default channel program limits
ZCP_DEFAULT_INSTRLIMIT = 10 * 1000 * 1000
ZCP_DEFAULT_MEMLIMIT = 10 * 1024 * 1024
#: Encryption wrapping key length
WRAPPING_KEY_LEN = 32
#: Encryption key location enum
zfs_key_location = enum(
    'ZFS_KEYLOCATION_NONE',
    'ZFS_KEYLOCATION_PROMPT',
    'ZFS_KEYLOCATION_URI'
)
#: Encryption key format enum
zfs_keyformat = enum(
    'ZFS_KEYFORMAT_NONE',
    'ZFS_KEYFORMAT_RAW',
    'ZFS_KEYFORMAT_HEX',
    'ZFS_KEYFORMAT_PASSPHRASE'
)
# Encryption algorithms enum
zio_encrypt = enum(
    'ZIO_CRYPT_INHERIT',
    'ZIO_CRYPT_ON',
    'ZIO_CRYPT_OFF',
    'ZIO_CRYPT_AES_128_CCM',
    'ZIO_CRYPT_AES_192_CCM',
    'ZIO_CRYPT_AES_256_CCM',
    'ZIO_CRYPT_AES_128_GCM',
    'ZIO_CRYPT_AES_192_GCM',
    'ZIO_CRYPT_AES_256_GCM'
)
# ZFS-specific error codes
zfs_errno = enum_with_offset(1024, [
        'ZFS_ERR_CHECKPOINT_EXISTS',
        'ZFS_ERR_DISCARDING_CHECKPOINT',
        'ZFS_ERR_NO_CHECKPOINT',
        'ZFS_ERR_DEVRM_IN_PROGRESS',
        'ZFS_ERR_VDEV_TOO_BIG',
        'ZFS_ERR_IOC_CMD_UNAVAIL',
        'ZFS_ERR_IOC_ARG_UNAVAIL',
        'ZFS_ERR_IOC_ARG_REQUIRED',
        'ZFS_ERR_IOC_ARG_BADTYPE',
        'ZFS_ERR_WRONG_PARENT',
        'ZFS_ERR_FROM_IVSET_GUID_MISSING',
        'ZFS_ERR_FROM_IVSET_GUID_MISMATCH',
        'ZFS_ERR_SPILL_BLOCK_FLAG_MISSING',
        'ZFS_ERR_UNKNOWN_SEND_STREAM_FEATURE',
        'ZFS_ERR_EXPORT_IN_PROGRESS',
        'ZFS_ERR_BOOKMARK_SOURCE_NOT_ANCESTOR',
        'ZFS_ERR_STREAM_TRUNCATED',
        'ZFS_ERR_STREAM_LARGE_BLOCK_MISMATCH',
        'ZFS_ERR_RESILVER_IN_PROGRESS',
        'ZFS_ERR_REBUILD_IN_PROGRESS',
        'ZFS_ERR_BADPROP',
        'ZFS_ERR_VDEV_NOTSUP',
        'ZFS_ERR_NOT_USER_NAMESPACE',
        'ZFS_ERR_RESUME_EXISTS',
        'ZFS_ERR_CRYPTO_NOTSUP',
        'ZFS_ERR_RAIDZ_EXPAND_IN_PROGRESS',
    ],
    {}
)
# compat before we used the enum helper for these values
ZFS_ERR_CHECKPOINT_EXISTS = zfs_errno.ZFS_ERR_CHECKPOINT_EXISTS
assert (ZFS_ERR_CHECKPOINT_EXISTS == 1024)
ZFS_ERR_DISCARDING_CHECKPOINT = zfs_errno.ZFS_ERR_DISCARDING_CHECKPOINT
ZFS_ERR_NO_CHECKPOINT = zfs_errno.ZFS_ERR_NO_CHECKPOINT
ZFS_ERR_DEVRM_IN_PROGRESS = zfs_errno.ZFS_ERR_DEVRM_IN_PROGRESS
ZFS_ERR_VDEV_TOO_BIG = zfs_errno.ZFS_ERR_VDEV_TOO_BIG
ZFS_ERR_WRONG_PARENT = zfs_errno.ZFS_ERR_WRONG_PARENT
ZFS_ERR_VDEV_NOTSUP = zfs_errno.ZFS_ERR_VDEV_NOTSUP
ZFS_ERR_RAIDZ_EXPAND_IN_PROGRESS = zfs_errno.ZFS_ERR_RAIDZ_EXPAND_IN_PROGRESS

# vim: softtabstop=4 tabstop=4 expandtab shiftwidth=4



# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2015 ClusterHQ
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Python bindings for ``libzfs_core``.
"""
from __future__ import absolute_import, division, print_function

CDEF = """

    enum lzc_send_flags {
        LZC_SEND_FLAG_EMBED_DATA = 1,
        LZC_SEND_FLAG_LARGE_BLOCK = 2,
        LZC_SEND_FLAG_COMPRESS = 4,
        LZC_SEND_FLAG_RAW = 8
    };

    typedef enum {
        DMU_OST_NONE,
        DMU_OST_META,
        DMU_OST_ZFS,
        DMU_OST_ZVOL,
        DMU_OST_OTHER,
        DMU_OST_ANY,
        DMU_OST_NUMTYPES
    } dmu_objset_type_t;

    #define MAXNAMELEN 256

    struct drr_begin {
        uint64_t drr_magic;
        uint64_t drr_versioninfo; /* was drr_version */
        uint64_t drr_creation_time;
        dmu_objset_type_t drr_type;
        uint32_t drr_flags;
        uint64_t drr_toguid;
        uint64_t drr_fromguid;
        char drr_toname[MAXNAMELEN];
    };

    typedef struct zio_cksum {
        uint64_t zc_word[4];
    } zio_cksum_t;

    typedef struct dmu_replay_record {
        enum {
            DRR_BEGIN, DRR_OBJECT, DRR_FREEOBJECTS,
            DRR_WRITE, DRR_FREE, DRR_END, DRR_WRITE_BYREF,
            DRR_SPILL, DRR_WRITE_EMBEDDED, DRR_NUMTYPES
        } drr_type;
        uint32_t drr_payloadlen;
        union {
            struct drr_begin drr_begin;
            /* ... */
            struct drr_checksum {
                uint64_t drr_pad[34];
                zio_cksum_t drr_checksum;
            } drr_checksum;
        } drr_u;
    } dmu_replay_record_t;

    typedef enum {
        DCP_CMD_NONE,
        DCP_CMD_RAW_RECV,
        DCP_CMD_NEW_KEY,
        DCP_CMD_INHERIT,
        DCP_CMD_FORCE_NEW_KEY,
        DCP_CMD_FORCE_INHERIT
    } dcp_cmd_t;

    int libzfs_core_init(void);
    void libzfs_core_fini(void);

    int lzc_bookmark(nvlist_t *, nvlist_t **);
    int lzc_change_key(const char *, uint64_t, nvlist_t *, uint8_t *, uint_t);
    int lzc_channel_program(const char *, const char *, uint64_t, uint64_t,
        nvlist_t *, nvlist_t **);
    int lzc_channel_program_nosync(const char *, const char *, uint64_t,
        uint64_t, nvlist_t *, nvlist_t **);
    int lzc_clone(const char *, const char *, nvlist_t *);
    int lzc_create(const char *, dmu_objset_type_t, nvlist_t *, uint8_t *,
        uint_t);
    int lzc_destroy_bookmarks(nvlist_t *, nvlist_t **);
    int lzc_destroy_snaps(nvlist_t *, boolean_t, nvlist_t **);
    boolean_t lzc_exists(const char *);
    int lzc_get_bookmarks(const char *, nvlist_t *, nvlist_t **);
    int lzc_get_holds(const char *, nvlist_t **);
    int lzc_hold(nvlist_t *, int, nvlist_t **);
    int lzc_load_key(const char *, boolean_t, uint8_t *, uint_t);
    int lzc_promote(const char *, nvlist_t *, nvlist_t **);
    int lzc_receive(const char *, nvlist_t *, const char *, boolean_t,
        boolean_t, int);
    int lzc_receive_one(const char *, nvlist_t *, const char *, boolean_t,
        boolean_t, boolean_t, int, const dmu_replay_record_t *, int,
        uint64_t *, uint64_t *, uint64_t *, nvlist_t **);
    int lzc_receive_resumable(const char *, nvlist_t *, const char *,
        boolean_t, boolean_t, int);
    int lzc_receive_with_cmdprops(const char *, nvlist_t *, nvlist_t *,
        uint8_t *, uint_t, const char *, boolean_t, boolean_t,
        boolean_t, int, const dmu_replay_record_t *, int, uint64_t *,
        uint64_t *, uint64_t *, nvlist_t **);
    int lzc_receive_with_heal(const char *, nvlist_t *, nvlist_t *,
        uint8_t *, uint_t, const char *, boolean_t, boolean_t, boolean_t,
        boolean_t, int, const dmu_replay_record_t *, int, uint64_t *,
        uint64_t *, uint64_t *, nvlist_t **);
    int lzc_receive_with_header(const char *, nvlist_t *, const char *,
        boolean_t, boolean_t, boolean_t, int, const dmu_replay_record_t *);
    int lzc_release(nvlist_t *, nvlist_t **);
    int lzc_reopen(const char *, boolean_t);
    int lzc_rollback(const char *, char *, int);
    int lzc_rollback_to(const char *, const char *);
    int lzc_send(const char *, const char *, int, enum lzc_send_flags);
    int lzc_send_resume(const char *, const char *, int, enum lzc_send_flags,
        uint64_t, uint64_t);
    int lzc_send_space(const char *, const char *, enum lzc_send_flags,
        uint64_t *);
    int lzc_snaprange_space(const char *, const char *, uint64_t *);
    int lzc_snapshot(nvlist_t *, nvlist_t *, nvlist_t **);
    int lzc_sync(const char *, nvlist_t *, nvlist_t **);
    int lzc_unload_key(const char *);
    int lzc_pool_checkpoint(const char *);
    int lzc_pool_checkpoint_discard(const char *);
    int lzc_rename(const char *, const char *);
    int lzc_destroy(const char *fsname);

    int lzc_inherit(const char *fsname, const char *name, nvlist_t *);
    int lzc_set_props(const char *, nvlist_t *, nvlist_t *, nvlist_t *);
    int lzc_list (const char *, nvlist_t *);
"""

SOURCE = """
#include <libzfs/libzfs_core.h>
"""

LIBRARY = "zfs_core"

# vim: softtabstop=4 tabstop=4 expandtab shiftwidth=4



# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2015 ClusterHQ
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Python bindings for ``libnvpair``.
"""
from __future__ import absolute_import, division, print_function

CDEF = """
    typedef ... nvlist_t;
    typedef ... nvpair_t;


    typedef enum {
        DATA_TYPE_UNKNOWN = 0,
        DATA_TYPE_BOOLEAN,
        DATA_TYPE_BYTE,
        DATA_TYPE_INT16,
        DATA_TYPE_UINT16,
        DATA_TYPE_INT32,
        DATA_TYPE_UINT32,
        DATA_TYPE_INT64,
        DATA_TYPE_UINT64,
        DATA_TYPE_STRING,
        DATA_TYPE_BYTE_ARRAY,
        DATA_TYPE_INT16_ARRAY,
        DATA_TYPE_UINT16_ARRAY,
        DATA_TYPE_INT32_ARRAY,
        DATA_TYPE_UINT32_ARRAY,
        DATA_TYPE_INT64_ARRAY,
        DATA_TYPE_UINT64_ARRAY,
        DATA_TYPE_STRING_ARRAY,
        DATA_TYPE_HRTIME,
        DATA_TYPE_NVLIST,
        DATA_TYPE_NVLIST_ARRAY,
        DATA_TYPE_BOOLEAN_VALUE,
        DATA_TYPE_INT8,
        DATA_TYPE_UINT8,
        DATA_TYPE_BOOLEAN_ARRAY,
        DATA_TYPE_INT8_ARRAY,
        DATA_TYPE_UINT8_ARRAY
    } data_type_t;
    typedef enum { B_FALSE, B_TRUE } boolean_t;

    typedef unsigned char uchar_t;
    typedef unsigned int uint_t;

    int nvlist_alloc(nvlist_t **, uint_t, int);
    void nvlist_free(nvlist_t *);

    int nvlist_unpack(char *, size_t, nvlist_t **, int);

    void dump_nvlist(nvlist_t *, int);
    int nvlist_dup(nvlist_t *, nvlist_t **, int);

    int nvlist_add_boolean(nvlist_t *, const char *);
    int nvlist_add_boolean_value(nvlist_t *, const char *, boolean_t);
    int nvlist_add_byte(nvlist_t *, const char *, uchar_t);
    int nvlist_add_int8(nvlist_t *, const char *, int8_t);
    int nvlist_add_uint8(nvlist_t *, const char *, uint8_t);
    int nvlist_add_int16(nvlist_t *, const char *, int16_t);
    int nvlist_add_uint16(nvlist_t *, const char *, uint16_t);
    int nvlist_add_int32(nvlist_t *, const char *, int32_t);
    int nvlist_add_uint32(nvlist_t *, const char *, uint32_t);
    int nvlist_add_int64(nvlist_t *, const char *, int64_t);
    int nvlist_add_uint64(nvlist_t *, const char *, uint64_t);
    int nvlist_add_string(nvlist_t *, const char *, const char *);
    int nvlist_add_nvlist(nvlist_t *, const char *, nvlist_t *);
    int nvlist_add_boolean_array(nvlist_t *, const char *,
        const boolean_t *, uint_t);
    int nvlist_add_byte_array(nvlist_t *, const char *,
        const uchar_t *, uint_t);
    int nvlist_add_int8_array(nvlist_t *, const char *,
        const int8_t *, uint_t);
    int nvlist_add_uint8_array(nvlist_t *, const char *,
        const uint8_t *, uint_t);
    int nvlist_add_int16_array(nvlist_t *, const char *,
        const int16_t *, uint_t);
    int nvlist_add_uint16_array(nvlist_t *, const char *,
        const uint16_t *, uint_t);
    int nvlist_add_int32_array(nvlist_t *, const char *,
        const int32_t *, uint_t);
    int nvlist_add_uint32_array(nvlist_t *, const char *,
        const uint32_t *, uint_t);
    int nvlist_add_int64_array(nvlist_t *, const char *,
        const int64_t *, uint_t);
    int nvlist_add_uint64_array(nvlist_t *, const char *,
        const uint64_t *, uint_t);
    int nvlist_add_string_array(nvlist_t *, const char *,
        const char * const *, uint_t);
    int nvlist_add_nvlist_array(nvlist_t *, const char *,
        const nvlist_t * const *, uint_t);

    nvpair_t *nvlist_next_nvpair(nvlist_t *, nvpair_t *);
    nvpair_t *nvlist_prev_nvpair(nvlist_t *, nvpair_t *);
    char *nvpair_name(nvpair_t *);
    data_type_t nvpair_type(nvpair_t *);
    int nvpair_type_is_array(nvpair_t *);
    int nvpair_value_boolean_value(nvpair_t *, boolean_t *);
    int nvpair_value_byte(nvpair_t *, uchar_t *);
    int nvpair_value_int8(nvpair_t *, int8_t *);
    int nvpair_value_uint8(nvpair_t *, uint8_t *);
    int nvpair_value_int16(nvpair_t *, int16_t *);
    int nvpair_value_uint16(nvpair_t *, uint16_t *);
    int nvpair_value_int32(nvpair_t *, int32_t *);
    int nvpair_value_uint32(nvpair_t *, uint32_t *);
    int nvpair_value_int64(nvpair_t *, int64_t *);
    int nvpair_value_uint64(nvpair_t *, uint64_t *);
    int nvpair_value_string(nvpair_t *, char **);
    int nvpair_value_nvlist(nvpair_t *, nvlist_t **);
    int nvpair_value_boolean_array(nvpair_t *, boolean_t **, uint_t *);
    int nvpair_value_byte_array(nvpair_t *, uchar_t **, uint_t *);
    int nvpair_value_int8_array(nvpair_t *, int8_t **, uint_t *);
    int nvpair_value_uint8_array(nvpair_t *, uint8_t **, uint_t *);
    int nvpair_value_int16_array(nvpair_t *, int16_t **, uint_t *);
    int nvpair_value_uint16_array(nvpair_t *, uint16_t **, uint_t *);
    int nvpair_value_int32_array(nvpair_t *, int32_t **, uint_t *);
    int nvpair_value_uint32_array(nvpair_t *, uint32_t **, uint_t *);
    int nvpair_value_int64_array(nvpair_t *, int64_t **, uint_t *);
    int nvpair_value_uint64_array(nvpair_t *, uint64_t **, uint_t *);
    int nvpair_value_string_array(nvpair_t *, char ***, uint_t *);
    int nvpair_value_nvlist_array(nvpair_t *, nvlist_t ***, uint_t *);
"""

SOURCE = """
#include <libzfs/sys/nvpair.h>
"""

LIBRARY = "nvpair"

# vim: softtabstop=4 tabstop=4 expandtab shiftwidth=4



# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2015 ClusterHQ
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Utility functions for casting to a specific C type.
"""
from __future__ import absolute_import, division, print_function

from .bindings.libnvpair import ffi as _ffi


def _ffi_cast(type_name):
    type_info = _ffi.typeof(type_name)

    def _func(value):
        # this is for overflow / underflow checking only
        if type_info.kind == 'enum':
            try:
                type_info.elements[value]
            except KeyError as e:
                raise OverflowError('Invalid enum <%s> value %s: %s' %
                                    (type_info.cname, value, e))
        else:
            _ffi.new(type_name + '*', value)
        return _ffi.cast(type_name, value)
    _func.__name__ = type_name
    return _func


uint8_t = _ffi_cast('uint8_t')
int8_t = _ffi_cast('int8_t')
uint16_t = _ffi_cast('uint16_t')
int16_t = _ffi_cast('int16_t')
uint32_t = _ffi_cast('uint32_t')
int32_t = _ffi_cast('int32_t')
uint64_t = _ffi_cast('uint64_t')
int64_t = _ffi_cast('int64_t')
boolean_t = _ffi_cast('boolean_t')
uchar_t = _ffi_cast('uchar_t')


# First element of the value tuple is a suffix for a single value function
# while the second element is for an array function
_type_to_suffix = {
    _ffi.typeof('uint8_t'):     ('uint8', 'uint8'),
    _ffi.typeof('int8_t'):      ('int8', 'int8'),
    _ffi.typeof('uint16_t'):    ('uint16', 'uint16'),
    _ffi.typeof('int16_t'):     ('int16', 'int16'),
    _ffi.typeof('uint32_t'):    ('uint32', 'uint32'),
    _ffi.typeof('int32_t'):     ('int32', 'int32'),
    _ffi.typeof('uint64_t'):    ('uint64', 'uint64'),
    _ffi.typeof('int64_t'):     ('int64', 'int64'),
    _ffi.typeof('boolean_t'):   ('boolean_value', 'boolean'),
    _ffi.typeof('uchar_t'):     ('byte', 'byte'),
}


# vim: softtabstop=4 tabstop=4 expandtab shiftwidth=4



# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2015 ClusterHQ
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Tests for _nvlist module.
The tests convert from a `dict` to C ``nvlist_t`` and back to a `dict`
and verify that no information is lost and value types are correct.
The tests also check that various error conditions like unsupported
value types or out of bounds values are detected.
"""
from __future__ import absolute_import, division, print_function

import unittest

from .._nvlist import nvlist_in, nvlist_out, _lib
from ..ctypes import (
    uint8_t, int8_t, uint16_t, int16_t, uint32_t, int32_t,
    uint64_t, int64_t, boolean_t, uchar_t
)


class TestNVList(unittest.TestCase):

    def _dict_to_nvlist_to_dict(self, props):
        res = {}
        nv_in = nvlist_in(props)
        with nvlist_out(res) as nv_out:
            _lib.nvlist_dup(nv_in, nv_out, 0)
        return res

    def _assertIntDictsEqual(self, dict1, dict2):
        self.assertEqual(
            len(dict1), len(dict1),
            b"resulting dictionary is of different size")
        for key in dict1.keys():
            self.assertEqual(int(dict1[key]), int(dict2[key]))

    def _assertIntArrayDictsEqual(self, dict1, dict2):
        self.assertEqual(
            len(dict1), len(dict1),
            b"resulting dictionary is of different size")
        for key in dict1.keys():
            val1 = dict1[key]
            val2 = dict2[key]
            self.assertEqual(
                len(val1), len(val2), b"array values of different sizes")
            for x, y in zip(val1, val2):
                self.assertEqual(int(x), int(y))

    def test_empty(self):
        res = self._dict_to_nvlist_to_dict({})
        self.assertEqual(len(res), 0, b"expected empty dict")

    def test_invalid_key_type(self):
        with self.assertRaises(TypeError):
            self._dict_to_nvlist_to_dict({1: None})

    def test_invalid_val_type__tuple(self):
        with self.assertRaises(TypeError):
            self._dict_to_nvlist_to_dict({b"key": (1, 2)})

    def test_invalid_val_type__set(self):
        with self.assertRaises(TypeError):
            self._dict_to_nvlist_to_dict({b"key": set(1, 2)})

    def test_invalid_array_val_type(self):
        with self.assertRaises(TypeError):
            self._dict_to_nvlist_to_dict({b"key": [(1, 2), (3, 4)]})

    def test_invalid_array_of_arrays_val_type(self):
        with self.assertRaises(TypeError):
            self._dict_to_nvlist_to_dict({b"key": [[1, 2], [3, 4]]})

    def test_string_value(self):
        props = {b"key": b"value"}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_implicit_boolean_value(self):
        props = {b"key": None}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_boolean_values(self):
        props = {b"key1": True, b"key2": False}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_explicit_boolean_true_value(self):
        props = {b"key": boolean_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_boolean_false_value(self):
        props = {b"key": boolean_t(0)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_boolean_invalid_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": boolean_t(2)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_boolean_another_invalid_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": boolean_t(-1)}
            self._dict_to_nvlist_to_dict(props)

    def test_uint64_value(self):
        props = {b"key": 1}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_uint64_max_value(self):
        props = {b"key": 2 ** 64 - 1}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_uint64_too_large_value(self):
        props = {b"key": 2 ** 64}
        with self.assertRaises(OverflowError):
            self._dict_to_nvlist_to_dict(props)

    def test_uint64_negative_value(self):
        props = {b"key": -1}
        with self.assertRaises(OverflowError):
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_uint64_value(self):
        props = {b"key": uint64_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_uint64_max_value(self):
        props = {b"key": uint64_t(2 ** 64 - 1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_uint64_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uint64_t(2 ** 64)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_uint64_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uint64_t(-1)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_uint32_value(self):
        props = {b"key": uint32_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_uint32_max_value(self):
        props = {b"key": uint32_t(2 ** 32 - 1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_uint32_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uint32_t(2 ** 32)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_uint32_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uint32_t(-1)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_uint16_value(self):
        props = {b"key": uint16_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_uint16_max_value(self):
        props = {b"key": uint16_t(2 ** 16 - 1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_uint16_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uint16_t(2 ** 16)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_uint16_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uint16_t(-1)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_uint8_value(self):
        props = {b"key": uint8_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_uint8_max_value(self):
        props = {b"key": uint8_t(2 ** 8 - 1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_uint8_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uint8_t(2 ** 8)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_uint8_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uint8_t(-1)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_byte_value(self):
        props = {b"key": uchar_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_byte_max_value(self):
        props = {b"key": uchar_t(2 ** 8 - 1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_byte_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uchar_t(2 ** 8)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_byte_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": uchar_t(-1)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_int64_value(self):
        props = {b"key": int64_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int64_max_value(self):
        props = {b"key": int64_t(2 ** 63 - 1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int64_min_value(self):
        props = {b"key": int64_t(-(2 ** 63))}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int64_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": int64_t(2 ** 63)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_int64_too_small_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": int64_t(-(2 ** 63) - 1)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_int32_value(self):
        props = {b"key": int32_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int32_max_value(self):
        props = {b"key": int32_t(2 ** 31 - 1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int32_min_value(self):
        props = {b"key": int32_t(-(2 ** 31))}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int32_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": int32_t(2 ** 31)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_int32_too_small_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": int32_t(-(2 ** 31) - 1)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_int16_value(self):
        props = {b"key": int16_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int16_max_value(self):
        props = {b"key": int16_t(2 ** 15 - 1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int16_min_value(self):
        props = {b"key": int16_t(-(2 ** 15))}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int16_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": int16_t(2 ** 15)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_int16_too_small_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": int16_t(-(2 ** 15) - 1)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_int8_value(self):
        props = {b"key": int8_t(1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int8_max_value(self):
        props = {b"key": int8_t(2 ** 7 - 1)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int8_min_value(self):
        props = {b"key": int8_t(-(2 ** 7))}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_explicit_int8_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": int8_t(2 ** 7)}
            self._dict_to_nvlist_to_dict(props)

    def test_explicit_int8_too_small_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": int8_t(-(2 ** 7) - 1)}
            self._dict_to_nvlist_to_dict(props)

    def test_nested_dict(self):
        props = {b"key": {}}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_nested_nested_dict(self):
        props = {b"key": {b"key": {}}}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_mismatching_values_array(self):
        props = {b"key": [1, b"string"]}
        with self.assertRaises(TypeError):
            self._dict_to_nvlist_to_dict(props)

    def test_mismatching_values_array2(self):
        props = {b"key": [True, 10]}
        with self.assertRaises(TypeError):
            self._dict_to_nvlist_to_dict(props)

    def test_mismatching_values_array3(self):
        props = {b"key": [1, False]}
        with self.assertRaises(TypeError):
            self._dict_to_nvlist_to_dict(props)

    def test_string_array(self):
        props = {b"key": [b"value", b"value2"]}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_boolean_array(self):
        props = {b"key": [True, False]}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_explicit_boolean_array(self):
        props = {b"key": [boolean_t(False), boolean_t(True)]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_uint64_array(self):
        props = {b"key": [0, 1, 2 ** 64 - 1]}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_uint64_array_too_large_value(self):
        props = {b"key": [0, 2 ** 64]}
        with self.assertRaises(OverflowError):
            self._dict_to_nvlist_to_dict(props)

    def test_uint64_array_negative_value(self):
        props = {b"key": [0, -1]}
        with self.assertRaises(OverflowError):
            self._dict_to_nvlist_to_dict(props)

    def test_mixed_explict_int_array(self):
        with self.assertRaises(TypeError):
            props = {b"key": [uint64_t(0), uint32_t(0)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_uint64_array(self):
        props = {b"key": [uint64_t(0), uint64_t(1), uint64_t(2 ** 64 - 1)]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_explict_uint64_array_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uint64_t(0), uint64_t(2 ** 64)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_uint64_array_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uint64_t(0), uint64_t(-1)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_uint32_array(self):
        props = {b"key": [uint32_t(0), uint32_t(1), uint32_t(2 ** 32 - 1)]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_explict_uint32_array_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uint32_t(0), uint32_t(2 ** 32)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_uint32_array_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uint32_t(0), uint32_t(-1)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_uint16_array(self):
        props = {b"key": [uint16_t(0), uint16_t(1), uint16_t(2 ** 16 - 1)]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_explict_uint16_array_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uint16_t(0), uint16_t(2 ** 16)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_uint16_array_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uint16_t(0), uint16_t(-1)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_uint8_array(self):
        props = {b"key": [uint8_t(0), uint8_t(1), uint8_t(2 ** 8 - 1)]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_explict_uint8_array_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uint8_t(0), uint8_t(2 ** 8)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_uint8_array_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uint8_t(0), uint8_t(-1)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_byte_array(self):
        props = {b"key": [uchar_t(0), uchar_t(1), uchar_t(2 ** 8 - 1)]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_explict_byte_array_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uchar_t(0), uchar_t(2 ** 8)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_byte_array_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [uchar_t(0), uchar_t(-1)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_int64_array(self):
        props = {b"key": [
            int64_t(0), int64_t(1), int64_t(2 ** 63 - 1), int64_t(-(2 ** 63))]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_explict_int64_array_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [int64_t(0), int64_t(2 ** 63)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_int64_array_too_small_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [int64_t(0), int64_t(-(2 ** 63) - 1)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_int32_array(self):
        props = {b"key": [
            int32_t(0), int32_t(1), int32_t(2 ** 31 - 1), int32_t(-(2 ** 31))]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_explict_int32_array_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [int32_t(0), int32_t(2 ** 31)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_int32_array_too_small_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [int32_t(0), int32_t(-(2 ** 31) - 1)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_int16_array(self):
        props = {b"key": [
            int16_t(0), int16_t(1), int16_t(2 ** 15 - 1), int16_t(-(2 ** 15))]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_explict_int16_array_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [int16_t(0), int16_t(2 ** 15)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_int16_array_too_small_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [int16_t(0), int16_t(-(2 ** 15) - 1)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_int8_array(self):
        props = {b"key": [
            int8_t(0), int8_t(1), int8_t(2 ** 7 - 1), int8_t(-(2 ** 7))]}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntArrayDictsEqual(props, res)

    def test_explict_int8_array_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [int8_t(0), int8_t(2 ** 7)]}
            self._dict_to_nvlist_to_dict(props)

    def test_explict_int8_array_too_small_value(self):
        with self.assertRaises(OverflowError):
            props = {b"key": [int8_t(0), int8_t(-(2 ** 7) - 1)]}
            self._dict_to_nvlist_to_dict(props)

    def test_dict_array(self):
        props = {b"key": [{b"key": 1}, {b"key": None}, {b"key": {}}]}
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)

    def test_implicit_uint32_value(self):
        props = {b"rewind-request": 1}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_implicit_uint32_max_value(self):
        props = {b"rewind-request": 2 ** 32 - 1}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_implicit_uint32_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"rewind-request": 2 ** 32}
            self._dict_to_nvlist_to_dict(props)

    def test_implicit_uint32_negative_value(self):
        with self.assertRaises(OverflowError):
            props = {b"rewind-request": -1}
            self._dict_to_nvlist_to_dict(props)

    def test_implicit_int32_value(self):
        props = {b"pool_context": 1}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_implicit_int32_max_value(self):
        props = {b"pool_context": 2 ** 31 - 1}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_implicit_int32_min_value(self):
        props = {b"pool_context": -(2 ** 31)}
        res = self._dict_to_nvlist_to_dict(props)
        self._assertIntDictsEqual(props, res)

    def test_implicit_int32_too_large_value(self):
        with self.assertRaises(OverflowError):
            props = {b"pool_context": 2 ** 31}
            self._dict_to_nvlist_to_dict(props)

    def test_implicit_int32_too_small_value(self):
        with self.assertRaises(OverflowError):
            props = {b"pool_context": -(2 ** 31) - 1}
            self._dict_to_nvlist_to_dict(props)

    def test_complex_dict(self):
        props = {
            b"key1": b"str",
            b"key2": 10,
            b"key3": {
                b"skey1": True,
                b"skey2": None,
                b"skey3": [
                    True,
                    False,
                    True
                ]
            },
            b"key4": [
                b"ab",
                b"bc"
            ],
            b"key5": [
                2 ** 64 - 1,
                1,
                2,
                3
            ],
            b"key6": [
                {
                    b"skey71": b"a",
                    b"skey72": b"b",
                },
                {
                    b"skey71": b"c",
                    b"skey72": b"d",
                },
                {
                    b"skey71": b"e",
                    b"skey72": b"f",
                }

            ],
            b"type": 2 ** 32 - 1,
            b"pool_context": -(2 ** 31)
        }
        res = self._dict_to_nvlist_to_dict(props)
        self.assertEqual(props, res)


# vim: softtabstop=4 tabstop=4 expandtab shiftwidth=4



# SPDX-License-Identifier: Apache-2.0
#
# Copyright 2015 ClusterHQ
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

"""
Helper routines for converting ``errno`` style error codes from C functions
to Python exceptions defined by `libzfs_core` API.

The conversion heavily depends on the context of the error: the attempted
operation and the input parameters.  For this reason, there is a conversion
routine for each `libzfs_core` interface function. The conversion routines
have the return code as a parameter as well as all the parameters of the
corresponding interface functions.

The parameters and exceptions are documented in the `libzfs_core` interfaces.
"""
from __future__ import absolute_import, division, print_function

import errno
import re
import string
from . import exceptions as lzc_exc
from ._constants import (
    ECHRNG,
    ECKSUM,
    ETIME,
    MAXNAMELEN,
    ZFS_ERR_CHECKPOINT_EXISTS,
    ZFS_ERR_DISCARDING_CHECKPOINT,
    ZFS_ERR_NO_CHECKPOINT,
    ZFS_ERR_DEVRM_IN_PROGRESS,
    ZFS_ERR_VDEV_TOO_BIG,
    ZFS_ERR_WRONG_PARENT,
    ZFS_ERR_RAIDZ_EXPAND_IN_PROGRESS,
    zfs_errno
)


def lzc_create_translate_error(ret, name, ds_type, props):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
        raise lzc_exc.PropertyInvalid(name)
    if ret == errno.EEXIST:
        raise lzc_exc.FilesystemExists(name)
    if ret == errno.ENOENT:
        raise lzc_exc.ParentNotFound(name)
    if ret == ZFS_ERR_WRONG_PARENT:
        raise lzc_exc.WrongParent(_fs_name(name))
    if ret == zfs_errno.ZFS_ERR_BADPROP:
        raise lzc_exc.PropertyInvalid(name)
    raise _generic_exception(ret, name, "Failed to create filesystem")


def lzc_clone_translate_error(ret, name, origin, props):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
        _validate_snap_name(origin)
        raise lzc_exc.PropertyInvalid(name)
    if ret == errno.EXDEV:
        raise lzc_exc.PoolsDiffer(name)
    if ret == errno.EEXIST:
        raise lzc_exc.FilesystemExists(name)
    if ret == errno.ENOENT:
        if not _is_valid_snap_name(origin):
            raise lzc_exc.SnapshotNameInvalid(origin)
        raise lzc_exc.DatasetNotFound(name)
    raise _generic_exception(ret, name, "Failed to create clone")


def lzc_rollback_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.ESRCH:
        raise lzc_exc.SnapshotNotFound(name)
    if ret == errno.EINVAL:
        _validate_fs_name(name)
        raise lzc_exc.NameInvalid(name)
    if ret == errno.ENOENT:
        if not _is_valid_fs_name(name):
            raise lzc_exc.NameInvalid(name)
        else:
            raise lzc_exc.FilesystemNotFound(name)
    raise _generic_exception(ret, name, "Failed to rollback")


def lzc_rollback_to_translate_error(ret, name, snap):
    if ret == errno.EEXIST:
        raise lzc_exc.SnapshotNotLatest(snap)
    else:
        lzc_rollback_translate_error(ret, name)


def lzc_snapshot_translate_errors(ret, errlist, snaps, props):
    if ret == 0:
        return

    def _map(ret, name):
        if ret == errno.EXDEV:
            pool_names = iter(map(_pool_name, snaps))
            pool_name = next(pool_names, None)
            same_pool = all(x == pool_name for x in pool_names)
            if same_pool:
                return lzc_exc.DuplicateSnapshots(name)
            else:
                return lzc_exc.PoolsDiffer(name)
        elif ret == errno.EINVAL:
            if any(not _is_valid_snap_name(s) for s in snaps):
                return lzc_exc.NameInvalid(name)
            elif any(len(s) > MAXNAMELEN for s in snaps):
                return lzc_exc.NameTooLong(name)
            else:
                return lzc_exc.PropertyInvalid(name)

        if ret == errno.EEXIST:
            return lzc_exc.SnapshotExists(name)
        if ret == errno.ENOENT:
            return lzc_exc.FilesystemNotFound(name)
        return _generic_exception(ret, name, "Failed to create snapshot")

    _handle_err_list(ret, errlist, snaps, lzc_exc.SnapshotFailure, _map)


def lzc_destroy_snaps_translate_errors(ret, errlist, snaps, defer):
    if ret == 0:
        return

    def _map(ret, name):
        if ret == errno.EEXIST:
            return lzc_exc.SnapshotIsCloned(name)
        if ret == errno.ENOENT:
            return lzc_exc.PoolNotFound(name)
        if ret == errno.EBUSY:
            return lzc_exc.SnapshotIsHeld(name)
        return _generic_exception(ret, name, "Failed to destroy snapshot")

    _handle_err_list(
        ret, errlist, snaps, lzc_exc.SnapshotDestructionFailure, _map)


def lzc_bookmark_translate_errors(ret, errlist, bookmarks):

    if ret == 0:
        return

    def _map(ret, name):
        source = bookmarks[name]
        if ret == errno.EINVAL:
            if name:
                pool_names = map(_pool_name, bookmarks.keys())

                # use _validate* functions for MAXNAMELEN check
                try:
                    _validate_bmark_name(name)
                except lzc_exc.ZFSError as e:
                    return e

                try:
                    _validate_snap_name(source)
                    source_is_snap = True
                except lzc_exc.ZFSError:
                    source_is_snap = False
                try:
                    _validate_bmark_name(source)
                    source_is_bmark = True
                except lzc_exc.ZFSError:
                    source_is_bmark = False
                if not source_is_snap and not source_is_bmark:
                    return lzc_exc.BookmarkSourceInvalid(source)

                if any(x != _pool_name(name) for x in pool_names):
                    return lzc_exc.PoolsDiffer(name)
            else:
                invalid_names = [
                    b for b in bookmarks.keys() if not _is_valid_bmark_name(b)]
                if invalid_names:
                    return lzc_exc.BookmarkNameInvalid(invalid_names[0])
        if ret == errno.EEXIST:
            return lzc_exc.BookmarkExists(name)
        if ret == errno.ENOENT:
            return lzc_exc.SnapshotNotFound(name)
        if ret == errno.ENOTSUP:
            return lzc_exc.BookmarkNotSupported(name)
        if ret == zfs_errno.ZFS_ERR_BOOKMARK_SOURCE_NOT_ANCESTOR:
            return lzc_exc.BookmarkMismatch(source)
        return _generic_exception(ret, name, "Failed to create bookmark")

    _handle_err_list(
        ret, errlist, bookmarks.keys(), lzc_exc.BookmarkFailure, _map)


def lzc_get_bookmarks_translate_error(ret, fsname, props):
    if ret == 0:
        return
    if ret == errno.ENOENT:
        raise lzc_exc.FilesystemNotFound(fsname)
    raise _generic_exception(ret, fsname, "Failed to list bookmarks")


def lzc_destroy_bookmarks_translate_errors(ret, errlist, bookmarks):
    if ret == 0:
        return

    def _map(ret, name):
        if ret == errno.EINVAL:
            return lzc_exc.NameInvalid(name)
        return _generic_exception(ret, name, "Failed to destroy bookmark")

    _handle_err_list(
        ret, errlist, bookmarks, lzc_exc.BookmarkDestructionFailure, _map)


def lzc_snaprange_space_translate_error(ret, firstsnap, lastsnap):
    if ret == 0:
        return
    if ret == errno.EXDEV and firstsnap is not None:
        if _pool_name(firstsnap) != _pool_name(lastsnap):
            raise lzc_exc.PoolsDiffer(lastsnap)
        else:
            raise lzc_exc.SnapshotMismatch(lastsnap)
    if ret == errno.EINVAL:
        if not _is_valid_snap_name(firstsnap):
            raise lzc_exc.NameInvalid(firstsnap)
        elif not _is_valid_snap_name(lastsnap):
            raise lzc_exc.NameInvalid(lastsnap)
        elif len(firstsnap) > MAXNAMELEN:
            raise lzc_exc.NameTooLong(firstsnap)
        elif len(lastsnap) > MAXNAMELEN:
            raise lzc_exc.NameTooLong(lastsnap)
        elif _pool_name(firstsnap) != _pool_name(lastsnap):
            raise lzc_exc.PoolsDiffer(lastsnap)
        else:
            raise lzc_exc.SnapshotMismatch(lastsnap)
    if ret == errno.ENOENT:
        raise lzc_exc.SnapshotNotFound(lastsnap)
    raise _generic_exception(
        ret, lastsnap, "Failed to calculate space used by range of snapshots")


def lzc_hold_translate_errors(ret, errlist, holds, fd):
    if ret == 0:
        return

    def _map(ret, name):
        if ret == errno.EXDEV:
            return lzc_exc.PoolsDiffer(name)
        elif ret == errno.EINVAL:
            if name:
                pool_names = map(_pool_name, holds.keys())
                if not _is_valid_snap_name(name):
                    return lzc_exc.NameInvalid(name)
                elif len(name) > MAXNAMELEN:
                    return lzc_exc.NameTooLong(name)
                elif any(x != _pool_name(name) for x in pool_names):
                    return lzc_exc.PoolsDiffer(name)
            else:
                invalid_names = [
                    b for b in holds.keys() if not _is_valid_snap_name(b)]
                if invalid_names:
                    return lzc_exc.NameInvalid(invalid_names[0])
        fs_name = None
        hold_name = None
        pool_name = None
        if name is not None:
            fs_name = _fs_name(name)
            pool_name = _pool_name(name)
            hold_name = holds[name]
        if ret == errno.ENOENT:
            return lzc_exc.FilesystemNotFound(fs_name)
        if ret == errno.EEXIST:
            return lzc_exc.HoldExists(name)
        if ret == errno.E2BIG:
            return lzc_exc.NameTooLong(hold_name)
        if ret == errno.ENOTSUP:
            return lzc_exc.FeatureNotSupported(pool_name)
        return _generic_exception(ret, name, "Failed to hold snapshot")

    if ret == errno.EBADF:
        raise lzc_exc.BadHoldCleanupFD()
    _handle_err_list(ret, errlist, holds.keys(), lzc_exc.HoldFailure, _map)


def lzc_release_translate_errors(ret, errlist, holds):
    if ret == 0:
        return
    for snap in holds:
        hold_list = holds[snap]
        if not isinstance(hold_list, list):
            raise lzc_exc.TypeError('holds must be in a list')

    def _map(ret, name):
        if ret == errno.EXDEV:
            return lzc_exc.PoolsDiffer(name)
        elif ret == errno.EINVAL:
            if name:
                pool_names = map(_pool_name, holds.keys())
                if not _is_valid_snap_name(name):
                    return lzc_exc.NameInvalid(name)
                elif len(name) > MAXNAMELEN:
                    return lzc_exc.NameTooLong(name)
                elif any(x != _pool_name(name) for x in pool_names):
                    return lzc_exc.PoolsDiffer(name)
            else:
                invalid_names = [
                    b for b in holds.keys() if not _is_valid_snap_name(b)]
                if invalid_names:
                    return lzc_exc.NameInvalid(invalid_names[0])
        elif ret == errno.ENOENT:
            return lzc_exc.HoldNotFound(name)
        elif ret == errno.E2BIG:
            tag_list = holds[name]
            too_long_tags = [t for t in tag_list if len(t) > MAXNAMELEN]
            return lzc_exc.NameTooLong(too_long_tags[0])
        elif ret == errno.ENOTSUP:
            pool_name = None
            if name is not None:
                pool_name = _pool_name(name)
            return lzc_exc.FeatureNotSupported(pool_name)
        else:
            return _generic_exception(
                ret, name, "Failed to release snapshot hold")

    _handle_err_list(
        ret, errlist, holds.keys(), lzc_exc.HoldReleaseFailure, _map)


def lzc_get_holds_translate_error(ret, snapname):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_snap_name(snapname)
    if ret == errno.ENOENT:
        raise lzc_exc.SnapshotNotFound(snapname)
    if ret == errno.ENOTSUP:
        raise lzc_exc.FeatureNotSupported(_pool_name(snapname))
    raise _generic_exception(ret, snapname, "Failed to get holds on snapshot")


def lzc_send_translate_error(ret, snapname, fromsnap, fd, flags):
    if ret == 0:
        return
    if ret == errno.EXDEV and fromsnap is not None:
        if _pool_name(fromsnap) != _pool_name(snapname):
            raise lzc_exc.PoolsDiffer(snapname)
        else:
            raise lzc_exc.SnapshotMismatch(snapname)
    elif ret == errno.EINVAL:
        if (fromsnap is not None and not _is_valid_snap_name(fromsnap) and
                not _is_valid_bmark_name(fromsnap)):
            raise lzc_exc.NameInvalid(fromsnap)
        elif (not _is_valid_snap_name(snapname) and
                not _is_valid_fs_name(snapname)):
            raise lzc_exc.NameInvalid(snapname)
        elif fromsnap is not None and len(fromsnap) > MAXNAMELEN:
            raise lzc_exc.NameTooLong(fromsnap)
        elif len(snapname) > MAXNAMELEN:
            raise lzc_exc.NameTooLong(snapname)
        elif (fromsnap is not None and
                _pool_name(fromsnap) != _pool_name(snapname)):
            raise lzc_exc.PoolsDiffer(snapname)
    elif ret == errno.ENOENT:
        if (fromsnap is not None and not _is_valid_snap_name(fromsnap) and
                not _is_valid_bmark_name(fromsnap)):
            raise lzc_exc.NameInvalid(fromsnap)
        raise lzc_exc.SnapshotNotFound(snapname)
    elif ret == errno.ENAMETOOLONG:
        if fromsnap is not None and len(fromsnap) > MAXNAMELEN:
            raise lzc_exc.NameTooLong(fromsnap)
        else:
            raise lzc_exc.NameTooLong(snapname)
    raise lzc_exc.StreamIOError(ret)


def lzc_send_space_translate_error(ret, snapname, fromsnap):
    if ret == 0:
        return
    if ret == errno.EXDEV and fromsnap is not None:
        if _pool_name(fromsnap) != _pool_name(snapname):
            raise lzc_exc.PoolsDiffer(snapname)
        else:
            raise lzc_exc.SnapshotMismatch(snapname)
    elif ret == errno.EINVAL:
        if fromsnap is not None and not _is_valid_snap_name(fromsnap):
            raise lzc_exc.NameInvalid(fromsnap)
        elif not _is_valid_snap_name(snapname):
            raise lzc_exc.NameInvalid(snapname)
        elif fromsnap is not None and len(fromsnap) > MAXNAMELEN:
            raise lzc_exc.NameTooLong(fromsnap)
        elif len(snapname) > MAXNAMELEN:
            raise lzc_exc.NameTooLong(snapname)
        elif (fromsnap is not None and
                _pool_name(fromsnap) != _pool_name(snapname)):
            raise lzc_exc.PoolsDiffer(snapname)
    elif ret == errno.ENOENT and fromsnap is not None:
        if not _is_valid_snap_name(fromsnap):
            raise lzc_exc.NameInvalid(fromsnap)
    if ret == errno.ENOENT:
        raise lzc_exc.SnapshotNotFound(snapname)
    raise _generic_exception(
        ret, snapname, "Failed to estimate backup stream size")


def lzc_receive_translate_errors(
    ret, snapname, fd, force, raw, resumable, embedded, origin, properrs
):
    if ret == 0:
        if properrs is not None and len(properrs) > 0:
            def _map(ret, name):
                if ret == errno.EINVAL:
                    return lzc_exc.PropertyInvalid(name)
                if ret == zfs_errno.ZFS_ERR_BADPROP:
                    return lzc_exc.PropertyInvalid(name)
                return _generic_exception(ret, name, "Failed to set property")
            _handle_err_list(
                errno.EINVAL, properrs, [snapname],
                lzc_exc.ReceivePropertyFailure, _map)
        else:
            return
    if ret == errno.EINVAL:
        if (not _is_valid_snap_name(snapname) and
                not _is_valid_fs_name(snapname)):
            raise lzc_exc.NameInvalid(snapname)
        elif len(snapname) > MAXNAMELEN:
            raise lzc_exc.NameTooLong(snapname)
        elif origin is not None and not _is_valid_snap_name(origin):
            raise lzc_exc.NameInvalid(origin)
        elif resumable:
            raise lzc_exc.StreamFeatureInvalid()
        elif embedded and not raw:
            raise lzc_exc.StreamFeatureIncompatible()
        else:
            raise lzc_exc.BadStream()
    if ret == errno.ENOENT:
        if not _is_valid_snap_name(snapname):
            raise lzc_exc.NameInvalid(snapname)
        else:
            raise lzc_exc.DatasetNotFound(snapname)
    if ret == errno.EEXIST:
        raise lzc_exc.DatasetExists(snapname)
    if ret == errno.ENOTSUP:
        raise lzc_exc.StreamFeatureNotSupported()
    if ret == errno.ENODEV:
        raise lzc_exc.StreamMismatch(_fs_name(snapname))
    if ret == errno.ETXTBSY:
        raise lzc_exc.DestinationModified(_fs_name(snapname))
    if ret == errno.EBUSY:
        raise lzc_exc.DatasetBusy(_fs_name(snapname))
    if ret == errno.ENOSPC:
        raise lzc_exc.NoSpace(_fs_name(snapname))
    if ret == errno.EDQUOT:
        raise lzc_exc.QuotaExceeded(_fs_name(snapname))
    if ret == errno.ENAMETOOLONG:
        raise lzc_exc.NameTooLong(snapname)
    if ret == errno.EROFS:
        raise lzc_exc.ReadOnlyPool(_pool_name(snapname))
    if ret == errno.EAGAIN:
        raise lzc_exc.SuspendedPool(_pool_name(snapname))
    if ret == errno.EACCES:
        raise lzc_exc.EncryptionKeyNotLoaded()
    if ret == ECKSUM:
        raise lzc_exc.BadStream()
    if ret == ZFS_ERR_WRONG_PARENT:
        raise lzc_exc.WrongParent(_fs_name(snapname))
    if ret == zfs_errno.ZFS_ERR_STREAM_TRUNCATED:
        raise lzc_exc.StreamTruncated()
    if ret == zfs_errno.ZFS_ERR_BADPROP:
        raise lzc_exc.PropertyInvalid(snapname)

    raise lzc_exc.StreamIOError(ret)


def lzc_promote_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
        raise lzc_exc.NotClone(name)
    if ret == errno.ENOTSOCK:
        raise lzc_exc.NotClone(name)
    if ret == errno.ENOENT:
        raise lzc_exc.FilesystemNotFound(name)
    if ret == errno.EEXIST:
        raise lzc_exc.SnapshotExists(name)
    raise _generic_exception(ret, name, "Failed to promote dataset")


def lzc_change_key_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
        raise lzc_exc.PropertyInvalid(name)
    if ret == errno.ENOENT:
        raise lzc_exc.FilesystemNotFound(name)
    if ret == errno.EACCES:
        raise lzc_exc.EncryptionKeyNotLoaded()
    raise _generic_exception(ret, name, "Failed to change encryption key")


def lzc_load_key_translate_error(ret, name, noop):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
        raise lzc_exc.PropertyInvalid(name)
    if ret == errno.ENOENT:
        raise lzc_exc.FilesystemNotFound(name)
    if ret == errno.EACCES:
        raise lzc_exc.EncryptionKeyInvalid()
    if ret == errno.EEXIST:
        raise lzc_exc.EncryptionKeyAlreadyLoaded()
    if noop:
        raise _generic_exception(ret, name, "Failed to load encryption key")
    else:
        raise _generic_exception(ret, name, "Failed to verify encryption key")


def lzc_unload_key_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
        raise lzc_exc.PropertyInvalid(name)
    if ret == errno.ENOENT:
        raise lzc_exc.FilesystemNotFound(name)
    if ret == errno.EACCES:
        raise lzc_exc.EncryptionKeyNotLoaded()
    raise _generic_exception(ret, name, "Failed to unload encryption key")


def lzc_sync_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.ENOENT:
        raise lzc_exc.PoolNotFound(name)
    raise _generic_exception(ret, name, "Failed to sync pool")


def lzc_reopen_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.ENOENT:
        raise lzc_exc.PoolNotFound(name)
    raise _generic_exception(ret, name, "Failed to reopen pool")


def lzc_channel_program_translate_error(ret, name, error):
    if ret == 0:
        return
    if ret == errno.ENOENT:
        raise lzc_exc.PoolNotFound(name)
    if ret == ETIME:
        raise lzc_exc.ZCPTimeout()
    if ret == errno.ENOMEM:
        raise lzc_exc.ZCPMemoryError()
    if ret == errno.ENOSPC:
        raise lzc_exc.ZCPSpaceError()
    if ret == errno.EPERM:
        raise lzc_exc.ZCPPermissionError()
    if ret == ECHRNG:
        raise lzc_exc.ZCPRuntimeError(error)
    if ret == errno.EINVAL:
        if error is None:
            raise lzc_exc.ZCPLimitInvalid()
        else:
            raise lzc_exc.ZCPSyntaxError(error)
    raise _generic_exception(ret, name, "Failed to execute channel program")


def lzc_pool_checkpoint_translate_error(ret, name, discard=False):
    if ret == 0:
        return
    if ret == errno.ENOENT:
        raise lzc_exc.PoolNotFound(name)
    if ret == ZFS_ERR_CHECKPOINT_EXISTS:
        raise lzc_exc.CheckpointExists()
    if ret == ZFS_ERR_NO_CHECKPOINT:
        raise lzc_exc.CheckpointNotFound()
    if ret == ZFS_ERR_DISCARDING_CHECKPOINT:
        raise lzc_exc.CheckpointDiscarding()
    if ret == ZFS_ERR_DEVRM_IN_PROGRESS:
        raise lzc_exc.DeviceRemovalRunning()
    if ret == ZFS_ERR_VDEV_TOO_BIG:
        raise lzc_exc.DeviceTooBig()
    if ret == ZFS_ERR_RAIDZ_EXPAND_IN_PROGRESS:
        raise lzc_exc.RaidzExpansionRunning()
    if discard:
        raise _generic_exception(
            ret, name, "Failed to discard pool checkpoint")
    else:
        raise _generic_exception(ret, name, "Failed to create pool checkpoint")


def lzc_pool_checkpoint_discard_translate_error(ret, name):
    lzc_pool_checkpoint_translate_error(ret, name, discard=True)


def lzc_rename_translate_error(ret, source, target):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(source)
        _validate_fs_name(target)
        if _pool_name(source) != _pool_name(target):
            raise lzc_exc.PoolsDiffer(source)
    if ret == errno.EEXIST:
        raise lzc_exc.FilesystemExists(target)
    if ret == errno.ENOENT:
        raise lzc_exc.FilesystemNotFound(source)
    if ret == ZFS_ERR_WRONG_PARENT:
        raise lzc_exc.WrongParent(target)
    raise _generic_exception(ret, source, "Failed to rename dataset")


def lzc_destroy_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
    if ret == errno.ENOENT:
        raise lzc_exc.FilesystemNotFound(name)
    raise _generic_exception(ret, name, "Failed to destroy dataset")


def lzc_inherit_prop_translate_error(ret, name, prop):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
        raise lzc_exc.PropertyInvalid(prop)
    if ret == errno.ENOENT:
        raise lzc_exc.DatasetNotFound(name)
    raise _generic_exception(ret, name, "Failed to inherit a property")


def lzc_set_prop_translate_error(ret, name, prop, val):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_or_snap_name(name)
        raise lzc_exc.PropertyInvalid(prop)
    if ret == errno.ENOENT:
        raise lzc_exc.DatasetNotFound(name)
    raise _generic_exception(ret, name, "Failed to set a property")


def lzc_get_props_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_or_snap_name(name)
    if ret == errno.ENOENT:
        raise lzc_exc.DatasetNotFound(name)
    raise _generic_exception(ret, name, "Failed to get properties")


def lzc_list_children_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
    raise _generic_exception(ret, name, "Error while iterating children")


def lzc_list_snaps_translate_error(ret, name):
    if ret == 0:
        return
    if ret == errno.EINVAL:
        _validate_fs_name(name)
    raise _generic_exception(ret, name, "Error while iterating snapshots")


def lzc_list_translate_error(ret, name, opts):
    if ret == 0:
        return
    if ret == errno.ENOENT:
        raise lzc_exc.DatasetNotFound(name)
    if ret == errno.EINVAL:
        _validate_fs_or_snap_name(name)
    raise _generic_exception(ret, name, "Error obtaining a list")


def _handle_err_list(ret, errlist, names, exception, mapper):
    '''
    Convert one or more errors from an operation into the requested exception.

    :param int ret: the overall return code.
    :param errlist: the dictionary that maps entity names to their specific
        error codes.
    :type errlist: dict of bytes:int
    :param names: the list of all names of the entities on which the operation
        was attempted.
    :param type exception: the type of the exception to raise if an error
        occurred. The exception should be a subclass of
        ``MultipleOperationsFailure``.
    :param function mapper: the function that maps an error code and a name to
        a Python exception.

    Unless ``ret`` is zero this function will raise the ``exception``.
    If the ``errlist`` is not empty, then the compound exception will contain
    a list of exceptions corresponding to each individual error code in the
    ``errlist``.
    Otherwise, the ``exception`` will contain a list with a single exception
    corresponding to the ``ret`` value. If the ``names`` list contains only one
    element, that is, the operation was attempted on a single entity, then the
    name of that entity is passed to the ``mapper``.
    If the operation was attempted on multiple entities, but the ``errlist``
    is empty, then we can not know which entity caused the error and, thus,
    ``None`` is used as a name to signify that fact.

    .. note::
        Note that the ``errlist`` can contain a special element with a key of
        "N_MORE_ERRORS".
        That element means that there were too many errors to place on the
        ``errlist``.
        Those errors are suppressed and only their count is provided as a
        value of the special ``N_MORE_ERRORS`` element.
    '''
    if ret == 0:
        return

    if len(errlist) == 0:
        suppressed_count = 0
        names = list(zip(names, range(2)))
        if len(names) == 1:
            name, _ = names[0]
        else:
            name = None
        errors = [mapper(ret, name)]
    else:
        errors = []
        suppressed_count = errlist.pop('N_MORE_ERRORS', 0)
        for name in errlist:
            err = errlist[name]
            errors.append(mapper(err, name))

    raise exception(errors, suppressed_count)


def _pool_name(name):
    '''
    Extract a pool name from the given dataset or bookmark name.

    '/' separates dataset name components.
    '@' separates a snapshot name from the rest of the dataset name.
    '#' separates a bookmark name from the rest of the dataset name.
    '''
    return re.split(b'[/@#]', name, 1)[0]


def _fs_name(name):
    '''
    Extract a dataset name from the given snapshot or bookmark name.

    '@' separates a snapshot name from the rest of the dataset name.
    '#' separates a bookmark name from the rest of the dataset name.
    '''
    return re.split(b'[@#]', name, 1)[0]


def _is_valid_name_component(component):
    allowed = string.ascii_letters + string.digits + u'-_.: '
    return component and all(x in allowed.encode() for x in component)


def _is_valid_fs_name(name):
    return name and all(_is_valid_name_component(c) for c in name.split(b'/'))


def _is_valid_snap_name(name):
    parts = name.split(b'@')
    return (len(parts) == 2 and _is_valid_fs_name(parts[0]) and
            _is_valid_name_component(parts[1]))


def _is_valid_bmark_name(name):
    parts = name.split(b'#')
    return (len(parts) == 2 and _is_valid_fs_name(parts[0]) and
            _is_valid_name_component(parts[1]))


def _validate_fs_name(name):
    if not _is_valid_fs_name(name):
        raise lzc_exc.FilesystemNameInvalid(name)
    elif len(name) > MAXNAMELEN:
        raise lzc_exc.NameTooLong(name)


def _validate_snap_name(name):
    if not _is_valid_snap_name(name):
        raise lzc_exc.SnapshotNameInvalid(name)
    elif len(name) > MAXNAMELEN:
        raise lzc_exc.NameTooLong(name)


def _validate_bmark_name(name):
    if not _is_valid_bmark_name(name):
        raise lzc_exc.BookmarkNameInvalid(name)
    elif len(name) > MAXNAMELEN:
        raise lzc_exc.NameTooLong(name)


def _validate_fs_or_snap_name(name):
    if not _is_valid_fs_name(name) and not _is_valid_snap_name(name):
        raise lzc_exc.NameInvalid(name)
    elif len(name) > MAXNAMELEN:
        raise lzc_exc.NameTooLong(name)


def _generic_exception(err, name, message):
    if err in _error_to_exception:
        return _error_to_exception[err](name)
    else:
        return lzc_exc.ZFSGenericError(err, message, name)


_error_to_exception = {e.errno: e for e in [
    lzc_exc.ZIOError,
    lzc_exc.NoSpace,
    lzc_exc.QuotaExceeded,
    lzc_exc.DatasetBusy,
    lzc_exc.NameTooLong,
    lzc_exc.ReadOnlyPool,
    lzc_exc.SuspendedPool,
    lzc_exc.PoolsDiffer,
    lzc_exc.PropertyNotSupported,
]}


# vim: softtabstop=4 tabstop=4 expandtab shiftwidth=4
